import gmsh

from typing import Any, Dict, Optional, List
from pydantic import BaseModel, validator
from meshwell.resolution import ResolutionSpec


class GMSH_entity(BaseModel):
    """
    Delayed evaluation of a gmsh occ kernel entity.

    Attributes:
        gmsh_function: entity-defining function from model.occ
        gmsh_function_kwargs: dict of keyword arguments for gmsh_function
        dim: dimension of the object (to properly generate dimtag)
        model: meshwell model
        physical_name: name of the physical this entity wil belong to
        mesh_order: priority of the entity if it overlaps with others (lower numbers override higher numbers)
        mesh_bool: if True, entity will be meshed; if not, will not be meshed (useful to tag boundaries)
    """

    gmsh_function: Any
    gmsh_function_kwargs: Dict[str, Any]
    dimension: int
    model: Any
    physical_name: Optional[str | tuple[str]] = None
    mesh_order: float | None = None
    mesh_bool: bool = True
    additive: bool = False
    resolutions: List[ResolutionSpec] | None = None

    @validator("physical_name", pre=True, always=True)
    def _set_physical_name(cls, physical_name: Optional[str | tuple[str]]):
        if isinstance(physical_name, str):
            return [physical_name]
        else:
            return physical_name

    def instanciate(self):
        """Returns dim tag from entity."""
        entity = self.gmsh_function(**self.gmsh_function_kwargs)
        self.model.occ.synchronize()
        return [(self.dimension, entity)]


if __name__ == "__main__":
    from meshwell.model import Model

    model = Model()

    gmsh_entity1 = GMSH_entity(
        gmsh_function=model.occ.add_box,
        gmsh_function_kwargs={"x": 0, "y": 0, "z": 0, "dx": 1, "dy": 1, "dz": 1},
        dimension=3,
        model=model,
    )

    gmsh_entity2 = GMSH_entity(
        gmsh_function=model.occ.addCylinder,
        gmsh_function_kwargs={
            "x": 2,
            "y": 2,
            "z": 2,
            "dx": 1,
            "dy": 0,
            "dz": 0,
            "r": 1,
            "angle": 1.0,
        },
        dimension=3,
        model=model,
    )

    tag1 = gmsh_entity1.instanciate()
    tag2 = gmsh_entity2.instanciate()
    print(tag1, tag2)

    model.model.mesh.generate(3)
    gmsh.write("mesh.msh")



from difflib import unified_diff
from pathlib import Path
from meshwell.config import PATH


def compare_meshes(meshfile: Path):
    meshfile1 = meshfile
    meshfile2 = PATH.references / (str(meshfile.with_suffix("")) + ".reference.msh")

    with open(str(meshfile1)) as f:
        expected_lines = f.readlines()
    with open(str(meshfile2)) as f:
        actual_lines = f.readlines()

    diff = list(unified_diff(expected_lines, actual_lines))
    assert diff == [], f"Mesh {str(meshfile)} differs from its reference!"



from __future__ import annotations

from os import cpu_count
from typing import Dict, Optional, List, Tuple
import numpy as np

from pathlib import Path

import gmsh
from meshwell.validation import (
    validate_dimtags,
    unpack_dimtags,
    order_entities,
    consolidate_entities_by_physical_name,
)
from meshwell.labeledentity import LabeledEntities
from meshwell.tag import tag_entities, tag_interfaces, tag_boundaries

import contextlib
import tempfile
import meshio


class Model:
    """Model class."""

    def __init__(
        self,
        point_tolerance: float = 1e-3,
        n_threads: int = cpu_count(),
    ):
        # Initialize model
        if gmsh.is_initialized():
            gmsh.clear()
        gmsh.initialize()

        # Set paralllel processing
        gmsh.option.setNumber("General.NumThreads", n_threads)
        gmsh.option.setNumber("Mesh.MaxNumThreads1D", n_threads)
        gmsh.option.setNumber("Mesh.MaxNumThreads2D", n_threads)
        gmsh.option.setNumber("Mesh.MaxNumThreads3D", n_threads)
        gmsh.option.setNumber("Geometry.OCCParallel", 1)

        # Point snapping
        self.point_tolerance = point_tolerance

        # CAD engine
        self.model = gmsh.model
        self.occ = self.model.occ

        # Track some gmsh entities for bottom-up volume definition
        self.points: Dict[Tuple[float, float, float], int] = {}
        self.segments: Dict[
            Tuple[Tuple[float, float, float], Tuple[float, float, float]], int
        ] = {}

    def add_get_point(self, x: float, y: float, z: float) -> int:
        """Add a point to the model, or reuse a previously-defined point.
        Args:
            x: float, x-coordinate
            y: float, y-coordinate
            z: float, z-coordinate
        Returns:
            ID of the added or retrieved point
        """
        if (x, y, z) not in self.points.keys():
            self.points[(x, y, z)] = self.occ.add_point(x, y, z)
        return self.points[(x, y, z)]

    def add_get_segment(
        self, xyz1: Tuple[float, float, float], xyz2: Tuple[float, float, float]
    ) -> int:
        """Add a segment (2-point line) to the gmsh model, or retrieve a previously-defined segment.
        The OCC kernel does not care about orientation.
        Args:
            xyz1: first [x,y,z] coordinate
            xyz2: second [x,y,z] coordinate
        Returns:
            ID of the added or retrieved line segment
        """
        if (xyz1, xyz2) in self.segments.keys():
            return self.segments[(xyz1, xyz2)]
        elif (xyz2, xyz1) in self.segments.keys():
            return self.segments[(xyz2, xyz1)]
        else:
            self.segments[(xyz1, xyz2)] = self.occ.add_line(
                self.add_get_point(xyz1[0], xyz1[1], xyz1[2]),
                self.add_get_point(xyz2[0], xyz2[1], xyz2[2]),
            )
            return self.segments[(xyz1, xyz2)]

    def _channel_loop_from_vertices(
        self, vertices: List[Tuple[float, float, float]]
    ) -> int:
        """Add a curve loop from the list of vertices.
        Args:
            vertices: list of [x,y,z] coordinates
        Returns:
            ID of the added curve loop
        """
        edges = []
        for vertex1, vertex2 in [
            (vertices[i], vertices[i + 1]) for i in range(len(vertices) - 1)
        ]:
            gmsh_line = self.add_get_segment(vertex1, vertex2)
            edges.append(gmsh_line)
        return self.occ.add_curve_loop(edges)

    def add_surface(self, vertices: List[Tuple[float, float, float]]) -> int:
        """Add a surface composed of the segments formed by vertices.

        Args:
            vertices: List of xyz coordinates, whose subsequent entries define a closed loop.
        Returns:
            ID of the added surface
        """
        channel_loop = self._channel_loop_from_vertices(vertices)
        try:
            return self.occ.add_plane_surface([channel_loop])
        except Exception as e:
            print("Failed vertices:")
            print(vertices)
            raise e

    def sync_model(self):
        """Synchronize the CAD model, and update points and lines vertex mapping."""
        self.occ.synchronize()
        cad_points = self.model.getEntities(dim=0)
        new_points: Dict[Tuple[float, float, float], int] = {}
        for _, cad_point in cad_points:
            # OCC kernel can do whatever, so just read point coordinates
            vertices = tuple(self.model.getValue(0, cad_point, []))
            new_points[vertices] = cad_point
        cad_lines = self.model.getEntities(dim=1)
        new_segments: Dict[
            Tuple[Tuple[float, float, float], Tuple[float, float, float]], int
        ] = {}
        for _, cad_line in cad_lines:
            try:
                key = list(self.segments.keys())[
                    list(self.segments.values()).index(cad_line)
                ]
                cad_lines[key] = cad_line
            except Exception:
                continue
        self.points = new_points
        self.segments = new_segments

    def mesh(
        self,
        entities_list: List,
        background_remeshing_file: Optional[Path] = None,
        default_characteristic_length: float = 0.5,
        global_scaling: float = 1.0,
        global_2D_algorithm: int = 6,
        global_3D_algorithm: int = 1,
        filename: Optional[str | Path] = None,
        verbosity: Optional[int] = 5,
        progress_bars: bool = True,
        interface_delimiter: str = "___",
        boundary_delimiter: str = "None",
        addition_delimiter: str = "+",
        addition_intersection_physicals: bool = True,
        addition_addition_physicals: bool = True,
        addition_structural_physicals: bool = True,
        gmsh_version: Optional[float] = None,
        finalize: bool = True,
        reinitialize: bool = True,
        periodic_entities: List[Tuple[str, str]] = None,
        fuse_entities_by_name: bool = False,
        optimization_flags: tuple[tuple[str, int]] | None = None,
    ) -> meshio.Mesh:
        """Creates a GMSH mesh with proper physical tagging from a dict of {labels: list( (GMSH entity dimension, GMSH entity tag) )}.

        Args:
            entities_list: list of meshwell entities (GMSH_entity, Prism, or PolySurface)
            background_remeshing (Path): path to a .pos file for background remeshing. If not None, is used instead of entity resolutions.
            default_characteristic_length (float): if resolutions is not specified for this physical, will use this value instead
            global_scaling: factor to scale all mesh coordinates by (e.g. 1E-6 to go from um to m)
            global_2D_algorithm: gmsh surface default meshing algorithm, see https://gmsh.info/doc/texinfo/gmsh.html#Mesh-options
            global_3D_algorithm: gmsh volume default meshing algorithm, see https://gmsh.info/doc/texinfo/gmsh.html#Mesh-options
            filename (str, path): if True, filepath where to save the mesh
            verbosity: GMSH stdout while meshing (True or False)
            interface_delimiter: string characters to use when naming interfaces between entities
            boundary_delimiter: string characters to use when defining an interface between an entity and nothing (simulation boundary)
            addition_delimiter: string characters to use with addition_intersection_physicals
            addition_addition_physicals: if True, the intersecting additive entities also carry the additive physical name
            addition_structural_physicals: if True, the intersecting additive entities also carry the underlying entity names
            addition_intersection_physicals: if True, the intersecting additive entities also carry a unique {underlying}{addition_delimiter}{additive} name
            gmsh_version: Gmsh mesh format version. For example, Palace requires an older version of 2.2,
                see https://mfem.org/mesh-formats/#gmsh-mesh-formats.
            finalize: if True (default), finalizes the GMSH model after execution
            periodic_entities: enforces mesh periodicity between the physical entities
            fuse_entities_by_name: if True, fuses CAD entities sharing the same physical_name
            optimization_flags: list of (method, niters) for smoothing. See https://gitlab.onelab.info/gmsh/gmsh/blob/gmsh_4_12_2/api/gmsh.py#L2087

        Returns:
            meshio object with mesh information
        """
        # Parse filename
        filename = str(filename)

        # If background mesh, create separate model
        if background_remeshing_file:
            # path = os.path.dirname(os.path.abspath(__file__))
            gmsh.merge(background_remeshing_file)
            gmsh.model.add("temp")

        gmsh.option.setNumber("General.Terminal", 10)  # 1 verbose, 0 otherwise
        gmsh.option.setNumber(
            "Mesh.CharacteristicLengthMax", default_characteristic_length
        )
        gmsh.option.setNumber("Mesh.Algorithm", global_2D_algorithm)
        gmsh.option.setNumber("Mesh.Algorithm3D", global_3D_algorithm)
        if gmsh_version is not None:
            gmsh.option.setNumber("Mesh.MshFileVersion", gmsh_version)

        # Initial synchronization
        self.occ.synchronize()

        # Parse and order the entities
        structural_entities = [
            entity for entity in entities_list if entity.additive is False
        ]
        additive_entities = [
            entity for entity in entities_list if entity.additive is True
        ]
        structural_entities = order_entities(structural_entities)
        additive_entities = order_entities(additive_entities)

        # Preserve ID numbering
        gmsh.option.setNumber("Geometry.OCCBooleanPreserveNumbering", 1)

        # Main loop:
        # Iterate through list of entities, generating and logging the volumes/surfaces in order
        # Manually remove intersections so that BooleanFragments from removeAllDuplicates does not reassign entity tags
        structural_entity_list = []
        max_dim = 0

        enumerator = enumerate(structural_entities)
        if progress_bars:
            from tqdm.auto import tqdm

            enumerator = tqdm(list(enumerator))

        for index, entity_obj in enumerator:
            physical_name = entity_obj.physical_name
            keep = entity_obj.mesh_bool
            resolutions = entity_obj.resolutions
            if progress_bars:
                if physical_name:
                    enumerator.set_description(
                        f"{str(physical_name):<30} - {'instanciate':<15}"
                    )
            # First create the shape
            dimtags_out = entity_obj.instanciate()

            if progress_bars:
                if physical_name:
                    enumerator.set_description(
                        f"{str(physical_name):<30} - {'dimtags':<15}"
                    )

            # Parse dimension
            dim = validate_dimtags(dimtags_out)
            max_dim = max(dim, max_dim)
            dimtags = unpack_dimtags(dimtags_out)

            if progress_bars:
                if physical_name:
                    enumerator.set_description(
                        f"{str(physical_name):<30} - {'entities':<15}"
                    )
            # Assemble with other shapes
            current_entities = LabeledEntities(
                index=index,
                dimtags=dimtags,
                physical_name=physical_name,
                keep=keep,
                model=self.model,
                resolutions=resolutions,
            )
            if progress_bars:
                if physical_name:
                    enumerator.set_description(
                        f"{str(physical_name):<30} - {'boolean':<15}"
                    )
            if index != 0:
                cut = self.occ.cut(
                    current_entities.dimtags,
                    [
                        dimtag
                        for previous_entities in structural_entity_list
                        for dimtag in previous_entities.dimtags
                    ],
                    removeObject=True,  # Only keep the difference
                    removeTool=False,  # Tool (previous entities) should remain untouched
                )
                # Heal interfaces now that there are no volume conflicts
                if progress_bars:
                    if physical_name:
                        enumerator.set_description(
                            f"{str(physical_name):<30} - {'duplicates':<15}"
                        )
                self.occ.removeAllDuplicates()
                if progress_bars:
                    if physical_name:
                        enumerator.set_description(
                            f"{str(physical_name):<30} - {'sync':<15}"
                        )
                self.sync_model()
                current_entities.dimtags = list(set(cut[0]))
            if current_entities.dimtags:
                structural_entity_list.append(current_entities)

        # Make sure the most up-to-date surfaces are logged as boundaries
        consolidated_entity_list = consolidate_entities_by_physical_name(
            structural_entity_list
        )
        structural_entity_list = []
        if fuse_entities_by_name:
            for entities in consolidated_entity_list:
                if len(entities.dimtags) != 1:
                    entities.dimtags = self.occ.fuse(
                        [entities.dimtags[0]],
                        entities.dimtags[1:],
                        removeObject=True,
                        removeTool=True,
                    )[0]
                    self.occ.synchronize()
                structural_entity_list.append(entities)
        else:
            structural_entity_list = consolidated_entity_list
        for entity in structural_entity_list:
            entity.update_boundaries()

        # Now that the structure is defined, merge with additive entities
        if additive_entities:
            for additive_entity in additive_entities:
                additive_dimtags = unpack_dimtags(additive_entity.instanciate())
                additive_entity_resolutions = (
                    []
                    if additive_entity.resolutions is None
                    else additive_entity.resolutions
                )
                updated_entities = []
                # Parse additive physical names
                for index, structural_entity in enumerate(structural_entity_list):
                    removeTool = False if index + 1 < len(structural_entities) else True

                    # Add all physical names
                    additive_names = []
                    if addition_addition_physicals:
                        additive_names.extend(additive_entity.physical_name)
                    if addition_structural_physicals:
                        additive_names.extend(structural_entity.physical_name)
                    if addition_intersection_physicals:
                        additive_names.extend(
                            [
                                f"{x}{addition_delimiter}{y}"
                                for x in structural_entity.physical_name
                                for y in additive_entity.physical_name
                            ]
                        )
                    additive_names = tuple(additive_names)

                    # Add the intersection and complement as new LabeledEntities
                    intersection = self.occ.intersect(
                        structural_entity.dimtags,
                        additive_dimtags,
                        removeObject=False,  # Don't remove yet
                        removeTool=removeTool,  # Only remove at the end
                    )
                    # No overlap case (no intersection)
                    if not intersection[0]:
                        self.occ.synchronize()
                        updated_entities.append(structural_entity)
                    else:
                        complement = self.occ.cut(
                            structural_entity.dimtags,
                            intersection[0],
                            removeObject=False,
                            removeTool=False,
                        )
                        self.occ.synchronize()
                        # Add the intersection and complement as new LabeledEntities
                        structural_entity_resolutions = (
                            []
                            if structural_entity.resolutions is None
                            else structural_entity.resolutions
                        )
                        if complement[0]:
                            updated_entities.append(
                                LabeledEntities(
                                    index=structural_entity.index,
                                    dimtags=complement[0],
                                    physical_name=structural_entity.physical_name,
                                    keep=structural_entity.keep,
                                    model=self.model,
                                    resolutions=structural_entity_resolutions,
                                )
                            )
                        updated_entities.append(
                            LabeledEntities(
                                index=structural_entity.index,
                                dimtags=intersection[0],
                                physical_name=additive_names,
                                keep=structural_entity.keep,
                                model=self.model,
                                resolutions=structural_entity_resolutions
                                + additive_entity_resolutions,
                            )
                        )
                # Use the updated entities in the next iteration
                structural_entity_list = updated_entities
                self.occ.removeAllDuplicates()
                self.sync_model()
        final_entity_list = structural_entity_list

        # Since we only took intersection with non-overlapping structural entities and removed the tool, there is no further conflict / tag reassignment
        # self.occ.removeAllDuplicates()
        for entity in final_entity_list:
            entity.update_boundaries()

        # Tag entities, interfaces, and boundaries
        tag_entities(final_entity_list)
        final_entity_list = tag_interfaces(
            final_entity_list, max_dim, interface_delimiter
        )
        tag_boundaries(
            final_entity_list, max_dim, interface_delimiter, boundary_delimiter
        )

        # Enforce periodic boundaries
        mapping = {}
        for dimtag in self.model.getPhysicalGroups():
            mapping[self.model.getPhysicalName(dimtag[0], dimtag[1])] = dimtag
        if periodic_entities:
            for label1, label2 in periodic_entities:
                if label1 not in mapping or label2 not in mapping:
                    continue
                tags1 = self.model.getEntitiesForPhysicalGroup(*mapping[label1])
                tags2 = self.model.getEntitiesForPhysicalGroup(*mapping[label2])

                vector1 = self.occ.getCenterOfMass(mapping[label1][0], tags1[0])
                vector2 = self.occ.getCenterOfMass(mapping[label1][0], tags2[0])
                vector = np.subtract(vector1, vector2)
                self.model.mesh.setPeriodic(
                    mapping[label1][0],
                    tags1,
                    tags2,
                    [
                        1,
                        0,
                        0,
                        vector[0],
                        0,
                        1,
                        0,
                        vector[1],
                        0,
                        0,
                        1,
                        vector[2],
                        0,
                        0,
                        0,
                        1,
                    ],
                )

        # Perform refinement
        if background_remeshing_file is None:
            # Use entity information
            refinement_field_indices = []
            refinement_max_index = 0
            for entity in final_entity_list:
                (
                    refinement_field_indices,
                    refinement_max_index,
                ) = entity.add_refinement_fields_to_model(
                    refinement_field_indices,
                    refinement_max_index,
                    default_characteristic_length,
                    final_entity_list,
                )

            # Use the smallest element size overall
            self.model.mesh.field.add("Min", refinement_max_index)
            self.model.mesh.field.setNumbers(
                refinement_max_index, "FieldsList", refinement_field_indices
            )
            self.model.mesh.field.setAsBackgroundMesh(refinement_max_index)
        else:
            bg_field = self.model.mesh.field.add("PostView")
            self.model.mesh.field.setNumber(bg_field, "ViewIndex", 0)
            gmsh.model.mesh.field.setAsBackgroundMesh(bg_field)

        # Remove boundary entities
        for entity in final_entity_list:
            if not entity.keep:
                self.model.occ.remove(entity.dimtags, recursive=True)

        # Turn off default meshing options
        gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
        gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)
        gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)

        # Global resizing
        gmsh.option.setNumber("Mesh.ScalingFactor", global_scaling)

        self.occ.synchronize()

        if not filename.endswith((".step", ".stp")):
            if global_3D_algorithm == 1 and verbosity:
                gmsh.logger.start()
            self.model.mesh.generate(max_dim)

            # Mesh smoothing
            if optimization_flags is not None:
                for optimization_flag, niter in optimization_flags:
                    self.model.mesh.optimize(optimization_flag, niter=niter)

            if global_3D_algorithm == 1 and verbosity:
                for line in gmsh.logger.get():
                    if "Optimizing volume " in str(line):
                        number = int(str(line).split("Optimizing volume ")[1])
                        physicalTags = gmsh.model.getPhysicalGroupsForEntity(3, number)
                        physicals = []
                        if len(physicalTags):
                            for p in physicalTags:
                                physicals.append(gmsh.model.getPhysicalName(dim, p))
                    if "ill-shaped tets are" in str(line):
                        print(",".join(physicals))
                        print(str(line))

        if filename:
            gmsh.write(f"{filename}")

        if not filename.endswith((".step", ".stp")):
            with contextlib.redirect_stdout(None):
                with tempfile.TemporaryDirectory() as tmpdirname:
                    gmsh.write(f"{tmpdirname}/mesh.msh")
                    if finalize:
                        gmsh.finalize()
                    return meshio.read(f"{tmpdirname}/mesh.msh")



from pydantic import BaseModel, Field, ConfigDict
from shapely.geometry import Polygon, MultiPolygon
from typing import List, Optional, Union, Any, Tuple
from meshwell.resolution import ResolutionSpec


class PolySurface(BaseModel):
    """
    Creates bottom-up GMSH polygonal surfaces formed by list of shapely (multi)polygon.

    Attributes:
        polygons: list of shapely (Multi)Polygon
        model: GMSH model to synchronize
        physical_name: name of the physical this entity wil belong to
        mesh_order: priority of the entity if it overlaps with others (lower numbers override higher numbers)
    """

    polygons: Union[Polygon, List[Polygon], MultiPolygon, List[MultiPolygon]] = Field(
        ...
    )
    model: Any
    physical_name: Optional[str | tuple[str, ...]] = Field(None)
    mesh_order: float | None = None
    mesh_bool: bool = Field(True)
    additive: bool = Field(False)
    dimension: int = Field(2)
    resolutions: List[ResolutionSpec] | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        polygons: Union[Polygon, List[Polygon], MultiPolygon, List[MultiPolygon]],
        model: Any,
        physical_name: Optional[str | tuple[str, ...]] = None,
        mesh_order: float | None = None,
        mesh_bool: bool = True,
        additive: bool = False,
        resolutions: List[ResolutionSpec] | None = None,
    ):
        super().__init__(
            polygons=polygons,
            model=model,
            physical_name=physical_name,
            mesh_order=mesh_order,
            mesh_bool=mesh_bool,
            additive=additive,
            resolution=resolutions,
        )

        # Parse (multi)polygons
        self.polygons = list(
            polygons.geoms if hasattr(polygons, "geoms") else [polygons]
        )

        # Track gmsh entities for bottom-up volume definition
        self.model = model

        self.mesh_order = mesh_order
        if isinstance(physical_name, str):
            self.physical_name = [physical_name]
        else:
            self.physical_name = physical_name
        self.mesh_bool = mesh_bool
        self.dimension = 2
        self.resolutions = resolutions
        self.additive = additive

    def _parse_coords(self, coords: Tuple[float, float]) -> Tuple[float, float, float]:
        """Chooses z=0 if the provided coordinates are 2D."""
        return (coords[0], coords[1], 0) if len(coords) == 2 else coords

    def get_gmsh_polygons(self) -> List[int]:
        """Returns the fused GMSH surfaces within model from the polygons."""
        surfaces = [self.add_surface_with_holes(entry) for entry in self.polygons]
        if len(surfaces) <= 1:
            return surfaces
        dimtags = self.model.occ.fuse(
            [(2, surfaces[0])],
            [(2, tag) for tag in surfaces[1:]],
            removeObject=True,
            removeTool=True,
        )[0]
        self.model.occ.synchronize()
        return [tag for dim, tag in dimtags]

    def add_surface_with_holes(self, polygon: Polygon) -> int:
        """Returns surface, removing intersection with hole surfaces."""
        exterior = self.model.add_surface(
            [self._parse_coords(coords) for coords in polygon.exterior.coords]
        )
        interior_tags = [
            self.model.add_surface(
                [self._parse_coords(coords) for coords in interior.coords],
            )
            for interior in polygon.interiors
        ]
        for interior_tag in interior_tags:
            exterior = self.model.occ.cut(
                [(2, exterior)], [(2, interior_tag)], removeObject=True, removeTool=True
            )
            self.model.occ.synchronize()
            exterior = exterior[0][0][1]  # Parse `outDimTags', `outDimTagsMap'
        return exterior

    def instanciate(self) -> List[Tuple[int, int]]:
        polysurface = self.get_gmsh_polygons()
        self.model.occ.synchronize()
        return [(2, polysurface)]



from pydantic import BaseModel, ConfigDict
import gmsh
from typing import List, Union, Any, Tuple
from meshwell.resolution import ResolutionSpec


class LabeledEntities(BaseModel):
    """General class to track the gmsh entities that result from the geometry definition."""

    index: int
    model: Any
    dimtags: List[Tuple[int, int]]
    physical_name: str | tuple[str, ...]
    resolutions: List[ResolutionSpec] | None = None
    keep: bool
    boundaries: List[int] = []
    interfaces: List = []

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def _fuse_self(self, dimtags: List[Union[int, str]]) -> List[Union[int, str]]:
        if len(dimtags) == 0:
            return []
        elif len(dimtags) != 1:
            dimtags = gmsh.model.occ.fuse(
                [dimtags[0]],
                dimtags[1:],
                removeObject=True,
                removeTool=True,
            )[0]
            self.model.occ.synchronize()
        return dimtags

    def get_tags(self) -> List[int]:
        tags = [tag for dim, tag in self.dimtags]
        if any(isinstance(el, list) for el in tags):
            tags = [item for sublist in tags for item in sublist]
        return tags

    def get_dim(self) -> int:
        return [dim for dim, tag in self.dimtags][0]

    def update_boundaries(self) -> List[int]:
        self.boundaries = [
            tag for dim, tag in gmsh.model.getBoundary(self.dimtags, True, False, False)
        ]
        return self.boundaries

    def add_refinement_fields_to_model(
        self,
        refinement_field_indices: List,
        refinement_max_index: int,
        default_resolution: float,
        final_entity_list: List,
    ):
        """
        Adds refinement fields to the model based on base_resolution and resolution info.

        # FIXME: make a better ResolutionSpec class that handles the entity/boundary distinction
        """
        n = refinement_max_index

        if self.resolutions:
            entities = []
            boundaries = []
            boundary_lines = []

            for resolutionspec in self.resolutions:
                # Parse by dimension
                if self.get_dim() == 3:
                    entity_str = "RegionsList"
                    boundary_str = "SurfacesList"
                    curves_str = "CurvesList"

                    # Check condition on volumes
                    if resolutionspec.resolution_volumes is None:
                        entities = []
                        entity_resolution = default_resolution
                    else:
                        entity_resolution = min(
                            resolutionspec.resolution_volumes, default_resolution
                        )
                        entities = [
                            volume_tag
                            for volume_tag in self.get_tags()
                            if resolutionspec.min_volumes
                            < self.model.occ.getMass(3, volume_tag)
                            < resolutionspec.max_volumes
                        ]

                    # Check condition on surfaces
                    if resolutionspec.resolution_surfaces is None:
                        boundaries = []
                        boundary_resolution = default_resolution
                    else:
                        boundaries = [
                            surface_tag
                            for surface_tag in self.boundaries
                            if resolutionspec.min_area_surfaces
                            < self.model.occ.getMass(2, surface_tag)
                            < resolutionspec.max_area_surfaces
                        ]
                        boundary_resolution = min(
                            resolutionspec.resolution_surfaces, entity_resolution
                        )
                        boundary_sizemax = resolutionspec.sizemax_surfaces
                        boundary_distmax = resolutionspec.distmax_surfaces
                        boundary_sigmoid = resolutionspec.surface_sigmoid
                        boundaries_samplings = {
                            boundary: resolutionspec.calculate_sampling_surface(
                                self.model.occ.getMass(2, boundary)
                            )
                            for boundary in boundaries
                        }

                    # Check condition on surface curves
                    if resolutionspec.resolution_curves is None:
                        boundary_lines = []
                    else:
                        boundary_lines = [
                            c
                            for b in self.boundaries
                            for cs in self.model.occ.getCurveLoops(b)[1]
                            for c in cs
                            if resolutionspec.min_length_curves
                            < self.model.occ.getMass(1, c)
                            < resolutionspec.max_length_curves
                        ]

                        boundary_lines_resolution = min(
                            resolutionspec.resolution_curves, boundary_resolution
                        )
                        boundary_lines_sizemax = resolutionspec.sizemax_curves
                        boundary_lines_distmax = resolutionspec.distmax_curves
                        boundary_line_sigmoid = resolutionspec.curve_sigmoid
                        boundary_lines_samplings = {
                            boundary_line: resolutionspec.calculate_sampling_curve(
                                self.model.occ.getMass(1, boundary_line)
                            )
                            for boundary_line in boundary_lines
                        }

                elif self.get_dim() == 2:
                    entity_str = "SurfacesList"
                    boundary_str = "CurvesList"
                    curves_str = "PointList"

                    # Check condition on surfaces
                    if resolutionspec.resolution_surfaces is None:
                        entities = []
                        entity_resolution = default_resolution
                    else:
                        entity_resolution = min(
                            resolutionspec.resolution_surfaces, default_resolution
                        )
                        entities = [
                            surface_tag
                            for surface_tag in self.get_tags()
                            if resolutionspec.min_area_surfaces
                            < self.model.occ.getMass(2, surface_tag)
                            < resolutionspec.max_area_surfaces
                        ]

                    # Check condition on curves
                    if resolutionspec.resolution_curves is None:
                        boundaries = []
                    else:
                        boundaries = [
                            curve_tag
                            for curve_tag in self.boundaries
                            if resolutionspec.min_length_curves
                            < self.model.occ.getMass(1, curve_tag)
                            < resolutionspec.max_length_curves
                        ]

                        boundary_resolution = min(
                            resolutionspec.resolution_curves, entity_resolution
                        )
                        boundary_sizemax = resolutionspec.sizemax_curves
                        boundary_distmax = resolutionspec.distmax_curves
                        boundary_sigmoid = resolutionspec.curve_sigmoid
                        boundaries_samplings = {
                            boundary: resolutionspec.calculate_sampling_curve(
                                self.model.occ.getMass(1, boundary)
                            )
                            for boundary in boundaries
                        }

                elif self.get_dim() == 1:
                    entity_str = "CurvesList"
                    boundary_str = "PointList"
                else:
                    entity_str = "PointList"
                    boundary_str = None

                if entities:
                    self.model.mesh.field.add("MathEval", n)
                    self.model.mesh.field.setString(n, "F", f"{entity_resolution}")
                    self.model.mesh.field.add("Restrict", n + 1)
                    self.model.mesh.field.setNumber(n + 1, "InField", n)
                    self.model.mesh.field.setNumbers(
                        n + 1,
                        entity_str,
                        entities,
                    )
                    refinement_field_indices.extend((n + 1,))
                    n += 2

                for boundary in boundaries:
                    self.model.mesh.field.add("Distance", n)
                    self.model.mesh.field.setNumbers(n, boundary_str, [boundary])
                    self.model.mesh.field.setNumber(
                        n, "Sampling", boundaries_samplings[boundary]
                    )
                    self.model.mesh.field.add("Threshold", n + 1)
                    self.model.mesh.field.setNumber(n + 1, "InField", n)
                    self.model.mesh.field.setNumber(
                        n + 1, "SizeMin", boundary_resolution
                    )
                    self.model.mesh.field.setNumber(n + 1, "DistMin", 0)
                    if boundary_sizemax and boundary_distmax:
                        self.model.mesh.field.setNumber(
                            n + 1, "SizeMax", boundary_sizemax
                        )
                        self.model.mesh.field.setNumber(
                            n + 1, "DistMax", boundary_distmax
                        )
                    self.model.mesh.field.setNumber(n + 1, "StopAtDistMax", 1)
                    self.model.mesh.field.setNumber(
                        n + 1, "Sigmoid", int(boundary_sigmoid)
                    )
                    refinement_field_indices.extend((n + 1,))
                    n += 2

                for boundary_line in boundary_lines:
                    self.model.mesh.field.add("Distance", n)
                    self.model.mesh.field.setNumbers(n, curves_str, [boundary_line])
                    self.model.mesh.field.setNumber(
                        n, "Sampling", boundary_lines_samplings[boundary_line]
                    )
                    self.model.mesh.field.add("Threshold", n + 1)
                    self.model.mesh.field.setNumber(n + 1, "InField", n)
                    self.model.mesh.field.setNumber(
                        n + 1, "SizeMin", boundary_lines_resolution
                    )
                    self.model.mesh.field.setNumber(n + 1, "DistMin", 0)
                    if boundary_lines_sizemax and boundary_lines_distmax:
                        self.model.mesh.field.setNumber(
                            n + 1, "SizeMax", boundary_lines_sizemax
                        )
                        self.model.mesh.field.setNumber(
                            n + 1, "DistMax", boundary_lines_distmax
                        )
                    self.model.mesh.field.setNumber(n + 1, "StopAtDistMax", 1)
                    self.model.mesh.field.setNumber(
                        n + 1, "Sigmoid", int(boundary_line_sigmoid)
                    )
                    refinement_field_indices.extend((n + 1,))
                    n += 2

        return refinement_field_indices, n



from typing import List
import gmsh
from itertools import combinations, product


def tag_entities(entity_list: List):
    """Adds physical labels to the entities in the model."""
    # One pass to get the global name --> dimtags mapping
    names_to_tags = {
        0: {},
        1: {},
        2: {},
        3: {},
    }
    for entities in entity_list:
        dim = entities.get_dim()
        for physical_name in entities.physical_name:
            if physical_name not in names_to_tags[dim]:
                names_to_tags[dim][physical_name] = []
            names_to_tags[dim][physical_name].extend(entities.get_tags())

    for dim in names_to_tags.keys():
        for physical_name, tags in names_to_tags[dim].items():
            gmsh.model.addPhysicalGroup(dim, tags, name=physical_name)


def tag_interfaces(entity_list: List, max_dim: int, boundary_delimiter: str):
    """Adds physical labels to the interfaces between entities in entity_list."""
    names_to_tags = {
        0: {},
        1: {},
        2: {},
    }
    for entity1, entity2 in combinations(entity_list, 2):
        if entity1.physical_name == entity2.physical_name:
            continue
        elif entity1.get_dim() != entity2.get_dim():
            continue
        elif entity1.get_dim() != max_dim:
            continue
        else:
            dim = entity1.get_dim() - 1
            common_interfaces = list(
                set(entity1.boundaries).intersection(entity2.boundaries)
            )
            if common_interfaces:
                # Update entity interface logs
                entity1.interfaces.extend(common_interfaces)
                entity2.interfaces.extend(common_interfaces)
                # Prepare physical tags
                for entity1_physical_name, entity2_physical_name in product(
                    entity1.physical_name, entity2.physical_name
                ):
                    interface_name = f"{entity1_physical_name}{boundary_delimiter}{entity2_physical_name}"
                    if interface_name not in names_to_tags[dim]:
                        names_to_tags[dim][interface_name] = []
                    names_to_tags[dim][interface_name].extend(common_interfaces)

    for dim in names_to_tags.keys():
        for physical_name, tags in names_to_tags[dim].items():
            gmsh.model.addPhysicalGroup(dim, tags, name=physical_name)

    return entity_list


def tag_boundaries(
    entity_list: List, max_dim: int, boundary_delimiter: str, mesh_edge_name: str
):
    """Adds physical labels to the boundaries of the entities in entity_list."""
    names_to_tags = {
        0: {},
        1: {},
        2: {},
    }
    for entity in entity_list:
        if entity.get_dim() != max_dim:
            continue
        dim = entity.get_dim() - 1
        boundaries = list(set(entity.boundaries) - set(entity.interfaces))
        for entity_physical_name in entity.physical_name:
            boundary_name = (
                f"{entity_physical_name}{boundary_delimiter}{mesh_edge_name}"
            )
            if boundary_name not in names_to_tags[dim]:
                names_to_tags[dim][boundary_name] = []
            names_to_tags[dim][boundary_name].extend(boundaries)

    for dim in names_to_tags.keys():
        for physical_name, tags in names_to_tags[dim].items():
            gmsh.model.addPhysicalGroup(dim, tags, name=physical_name)



import numpy as np
import copy
from pydantic import BaseModel


class ResolutionSpec(BaseModel):
    """
    Object holding resolution information for an entity and its boundaries.

    # FIXME: make a better ResolutionSpec class that handles the entity/boundary distinction

    Arguments:
        # Volume
        volume_resolution (float): resolution of the volume (3D). No effect if the entity is 2D.
            defaults to inf --> default resolution, since the local resolution is always the minimum of all size fields

        # Surface
        surface_resolution (float): resolution of the surface (2D) or of all the surfaces touching the volume (3D)
            defaults to inf --> default resolution (2D) or volume_resolution (3D), since the local resolution is always the minimum of all size fields

        # Curves
        curve_resolution (float): resolution of curves constituting the volumes' surfaces (3D) or surfaces (2D)
            defaults to inf --> surface_resolution, since the local resolution is always the minimum of all size fields

        # Points
        point_resolution (float): resolution of points constituting the volumes' surfaces' curves (3D) or surfaces' curves (2D)
            defaults to inf --> curve_resolution, since the local resolution is always the minimum of all size fields
    """

    # Volume
    resolution_volumes: float | None = None
    min_volumes: float = 0
    max_volumes: float = np.inf
    # Surface
    resolution_surfaces: float | None = None
    min_area_surfaces: float = 0
    max_area_surfaces: float = np.inf
    distmax_surfaces: float | None = None
    sizemax_surfaces: float | None = None
    surface_sigmoid: bool = False
    surface_per_sampling_surfaces: float | None = None
    sampling_surface_max: int = 100
    # Curve
    resolution_curves: float | None = None
    min_length_curves: float = 0
    max_length_curves: float = np.inf
    distmax_curves: float | None = None
    sizemax_curves: float | None = None
    curve_sigmoid: bool = False
    length_per_sampling_curves: float | None = None
    sampling_curve_max: int = 100
    # Point
    resolution_points: float | None = None
    distmax_points: float | None = None
    sizemax_points: float | None = None
    point_sigmoid: bool = False

    def refine(self, resolution_factor: float):
        result = copy.copy(self)
        if result.resolution_volumes is not None:
            result.resolution_volumes *= resolution_factor
        if result.resolution_surfaces is not None:
            result.resolution_surfaces *= resolution_factor
        if result.sizemax_surfaces is not None:
            result.sizemax_surfaces *= resolution_factor
        if result.resolution_curves is not None:
            result.resolution_curves *= resolution_factor
        if result.sizemax_curves is not None:
            result.sizemax_curves *= resolution_factor
        if result.resolution_points is not None:
            result.resolution_points *= resolution_factor
        if result.sizemax_points is not None:
            result.sizemax_points *= resolution_factor
        return result

    def calculate_sampling(self, mass_per_sampling, mass, max_sampling):
        if mass_per_sampling == np.inf:
            return 2  # avoid int(inf) error
        else:
            return min(max(2, int(mass / mass_per_sampling)), max_sampling)

    def calculate_sampling_surface(self, area):
        if self.surface_per_sampling_surfaces:
            return self.calculate_sampling(
                self.surface_per_sampling_surfaces, area, self.sampling_surface_max
            )
        else:
            return self.calculate_sampling(
                0.5 * self.resolution_surfaces, area, self.sampling_surface_max
            )

    def calculate_sampling_curve(self, length):
        if self.length_per_sampling_curves:
            return self.calculate_sampling(
                self.length_per_sampling_curves, length, self.sampling_curve_max
            )
        else:
            return self.calculate_sampling(
                0.5 * self.resolution_curves, length, self.sampling_curve_max
            )



from typing import List, Dict, Optional, Tuple, Union, Any
from pydantic import BaseModel, Field, ConfigDict
from shapely.geometry import Polygon, MultiPolygon
from meshwell.resolution import ResolutionSpec


class Prism(BaseModel):
    """
    Creates a bottom-up GMSH "prism" formed by a polygon associated with (optional) z-dependent grow/shrink operations.

    Attributes:
        polygons: list of shapely (Multi)Polygon
        buffers: dict of {z: buffer} used to shrink/grow base polygons at specified z-values
        physical_name: name of the physical this entity wil belong to
        mesh_order: priority of the entity if it overlaps with others (lower numbers override higher numbers)
        mesh_bool: if True, entity will be meshed; if not, will not be meshed (useful to tag boundaries)
    """

    polygons: Union[Polygon, List[Polygon], MultiPolygon, List[MultiPolygon]] = Field(
        ...
    )
    buffers: Dict[float, float] = Field(...)
    model: Any
    physical_name: Optional[str | tuple[str, ...]] = Field(None)
    mesh_order: float | None = None
    mesh_bool: bool = Field(True)
    additive: bool = Field(False)
    buffered_polygons: Optional[List[Tuple[float, Polygon]]] = []
    dimension: int = Field(3)
    resolutions: List[ResolutionSpec] | None = None
    extrude: bool = False
    zmin: Optional[float] = 0
    zmax: Optional[float] = 0
    subdivision: tuple[int, int, int] | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        polygons: Union[Polygon, List[Polygon], MultiPolygon, List[MultiPolygon]],
        buffers: Dict[float, float],
        model: Any,
        physical_name: Optional[str | tuple[str, ...]] = None,
        mesh_order: float | None = None,
        mesh_bool: bool = True,
        additive: bool = False,
        resolutions: List[ResolutionSpec] | None = None,
        subdivision: tuple[int, int, int] | None = None,
    ):
        super().__init__(
            polygons=polygons,
            buffers=buffers,
            model=model,
            physical_name=physical_name,
            mesh_order=mesh_order,
            mesh_bool=mesh_bool,
            additive=additive,
            resolution=resolutions,
            subdivision=subdivision,
        )

        # Model
        self.model = model

        # Parse buffers or prepare extrusion
        if all(buffer == 0 for buffer in buffers.values()):
            self.extrude = True
            self.polygons = polygons
            self.zmin, self.zmax = min(buffers.keys()), max(buffers.keys())
        else:
            self.extrude = False
            self.buffered_polygons: List[
                Tuple[float, Polygon]
            ] = self._get_buffered_polygons(polygons, buffers)

            # Validate the input
            if not self._validate_polygon_buffers():
                raise ValueError(
                    f"The buffer has modified the polygon vertices! Bad physical: {physical_name}"
                )

        # Mesh order and name
        self.mesh_order = mesh_order

        # Format physical name
        if isinstance(physical_name, str):
            self.physical_name = [physical_name]
        else:
            self.physical_name = physical_name
        self.mesh_bool = mesh_bool
        self.dimension = 3
        self.resolutions = resolutions
        self.additive = additive

    def get_gmsh_volumes(self) -> List[int]:
        """Returns the fused GMSH volumes within model from the polygons and buffers."""
        if self.extrude:
            surfaces = [
                (2, surface)
                for surface in self._add_surfaces_with_holes(self.polygons, self.zmin)
            ]
            entities = self.model.occ.extrude(surfaces, 0, 0, self.zmax - self.zmin)
            volumes = [tag for dim, tag in entities if dim == 3]
        else:
            volumes = [
                self._add_volume_with_holes(entry) for entry in self.buffered_polygons
            ]
        if len(volumes) <= 1:
            self.model.occ.synchronize()
            return volumes
        dimtags = self.model.occ.fuse(
            [(3, volumes[0])],
            [(3, tag) for tag in volumes[1:]],
            removeObject=True,
            removeTool=True,
        )[0]
        self.model.occ.synchronize()
        return [tag for dim, tag in dimtags]

    def _get_buffered_polygons(
        self, polygons: List[Polygon], buffers: Dict[float, float]
    ) -> List[Tuple[float, Polygon]]:
        """Break up polygons on each layer into lists of (z,polygon) tuples according to buffer entries.

        Arguments (implicit):
            polygons: list of (Multi)Polygons to bufferize
            buffers: {z: buffer} values to apply to the polygons

        Returns:
            buffered_polygons: list of (z, buffered_polygons)
        """
        all_polygons_list = []
        for polygon in polygons.geoms if hasattr(polygons, "geoms") else [polygons]:
            current_buffers = []
            for z, width_buffer in buffers.items():
                current_buffers.append((z, polygon.buffer(width_buffer, join_style=2)))
            all_polygons_list.append(current_buffers)

        return all_polygons_list

    def _add_volume(
        self,
        entry: List[Tuple[float, Polygon]],
        exterior: bool = True,
        interior_index: int = 0,
    ) -> int:
        """Create shape from a list of the same buffered polygon and a list of z-values.
        Args:
            polygons: shapely polygons from the GDS
            zs: list of z-values for each polygon
        Returns:
            ID of the added volume
        """
        # Draw bottom surface
        bottom_polygon = entry[0][1]
        bottom_z = entry[0][0]
        bottom_polygon_vertices = self.xy_surface_vertices(
            polygon=bottom_polygon,
            polygon_z=bottom_z,
            exterior=exterior,
            interior_index=interior_index,
        )
        gmsh_surfaces = [self.model.add_surface(bottom_polygon_vertices)]

        # Draw top surface
        top_polygon = entry[-1][1]
        top_z = entry[-1][0]
        top_polygon_vertices = self.xy_surface_vertices(
            polygon=top_polygon,
            polygon_z=top_z,
            exterior=exterior,
            interior_index=interior_index,
        )
        gmsh_surfaces.append(self.model.add_surface(top_polygon_vertices))

        # Draw vertical surfaces
        for pair_index in range(len(entry) - 1):
            if exterior:
                bottom_polygon = entry[pair_index][1].exterior.coords
                top_polygon = entry[pair_index + 1][1].exterior.coords
            else:
                bottom_polygon = entry[pair_index][1].interiors[interior_index].coords
                top_polygon = entry[pair_index + 1][1].interiors[interior_index].coords
            bottom_z = entry[pair_index][0]
            top_z = entry[pair_index + 1][0]
            for facet_pt_ind in range(len(bottom_polygon) - 1):
                facet_pt1 = (
                    bottom_polygon[facet_pt_ind][0],
                    bottom_polygon[facet_pt_ind][1],
                    bottom_z,
                )
                facet_pt2 = (
                    bottom_polygon[facet_pt_ind + 1][0],
                    bottom_polygon[facet_pt_ind + 1][1],
                    bottom_z,
                )
                facet_pt3 = (
                    top_polygon[facet_pt_ind + 1][0],
                    top_polygon[facet_pt_ind + 1][1],
                    top_z,
                )
                facet_pt4 = (
                    top_polygon[facet_pt_ind][0],
                    top_polygon[facet_pt_ind][1],
                    top_z,
                )
                facet_vertices = [facet_pt1, facet_pt2, facet_pt3, facet_pt4, facet_pt1]
                gmsh_surfaces.append(self.model.add_surface(facet_vertices))

        # Return volume from closed shell
        surface_loop = self.model.occ.add_surface_loop(gmsh_surfaces)
        return self.model.occ.add_volume([surface_loop])

    def xy_surface_vertices(
        self,
        polygon: Polygon,
        polygon_z: float,
        exterior: bool,
        interior_index: int,
    ) -> List[Tuple[float, float, float]]:
        """"""
        # Draw xy surface
        return (
            [(x, y, polygon_z) for x, y in polygon.exterior.coords]
            if exterior
            else [
                (x, y, polygon_z) for x, y in polygon.interiors[interior_index].coords
            ]
        )

    def _add_volume_with_holes(self, entry: List[Tuple[float, Polygon]]) -> int:
        """Returns volume, removing intersection with hole volumes."""
        exterior = self._add_volume(entry, exterior=True)
        interiors = [
            self._add_volume(
                entry,
                exterior=False,
                interior_index=interior_index,
            )
            for interior_index in range(len(entry[0][1].interiors))
        ]
        if interiors:
            exterior = self.model.occ.cut(
                [(3, exterior)],
                [(3, interior) for interior in interiors],
                removeObject=True,
                removeTool=True,
            )
            self.model.occ.synchronize()
            exterior = exterior[0][0][1]  # Parse `outDimTags', `outDimTagsMap'
        return exterior

    def subdivide(self, prisms, subdivision):
        """Split the prisms into subprisms according to subdivision."""
        subdivided_prisms = []
        import numpy as np

        global_xmin = np.inf
        global_ymin = np.inf
        global_zmin = np.inf
        global_xmax = -np.inf
        global_ymax = -np.inf
        global_zmax = -np.inf
        for prism in prisms:
            xmin, ymin, zmin, xmax, ymax, zmax = self.model.occ.getBoundingBox(3, prism)
            if xmin < global_xmin:
                global_xmin = xmin
            if ymin < global_ymin:
                global_ymin = ymin
            if zmin < global_zmin:
                global_zmin = zmin
            if xmax > global_xmax:
                global_xmax = xmax
            if ymax > global_ymax:
                global_ymax = ymax
            if zmax > global_zmax:
                global_zmax = zmax
        dx = (global_xmax - global_xmin) / subdivision[0]
        dy = (global_ymax - global_ymin) / subdivision[1]
        dz = (global_zmax - global_zmin) / subdivision[2]
        removeObject = False
        for x_index in range(subdivision[0]):
            for y_index in range(subdivision[1]):
                for z_index in range(subdivision[2]):
                    if (
                        x_index == subdivision[0] - 1
                        and y_index == subdivision[1] - 1
                        and z_index == subdivision[2] - 1
                    ):
                        removeObject = True
                    tool = self.model.occ.add_box(
                        xmin + x_index * dx,
                        ymin + y_index * dy,
                        zmin + z_index * dz,
                        dx,
                        dy,
                        dz,
                    )
                    intersection = self.model.occ.intersect(
                        [(3, prism) for prism in prisms],
                        [(3, tool)],
                        removeObject=removeObject,
                        removeTool=True,
                    )[0]
                    subdivided_prisms.extend(intersection)
        return subdivided_prisms

    def instanciate(self) -> List[Tuple[int, int]]:
        """Returns dim tag from entity."""
        prisms = self.get_gmsh_volumes()
        self.model.occ.synchronize()
        if self.subdivision is not None:
            return self.subdivide(prisms, self.subdivision)
        return [(3, prisms)]

    def _validate_polygon_buffers(self) -> bool:
        """Check if any buffering operation changes the topology of the polygon."""
        reference_exterior_vertices = len(
            self.buffered_polygons[0][0][1].exterior.coords
        )
        reference_interiors_vertices = [
            len(interior.coords)
            for interior in self.buffered_polygons[0][0][1].interiors
        ]
        for buffered_polygon in self.buffered_polygons[0][1:]:
            num_points_exterior = len(buffered_polygon[1].exterior.coords)
            num_points_interiors = (
                [
                    len(interior.coords) if interior else 0
                    for interior in buffered_polygon[1].interiors
                ]
                if buffered_polygon[1].interiors
                else [0]
            )
            if num_points_exterior != reference_exterior_vertices:
                return False
            for num_points_interior, reference_interior_vertices in zip(
                num_points_interiors, reference_interiors_vertices
            ):
                if num_points_interior != reference_interior_vertices:
                    return False
        return True

    """
    Extrusion method
    """

    def _add_surfaces_with_holes(self, polygons, z) -> List[int]:
        """Returns surface, removing intersection with hole surfaces."""
        surfaces = []
        for polygon in polygons.geoms if hasattr(polygons, "geoms") else [polygons]:
            # Add outer surface(s)
            exterior = self.model.add_surface(
                self.xy_surface_vertices(
                    polygon, polygon_z=z, exterior=True, interior_index=0
                )
            )
            interiors = [
                self.model.add_surface(
                    self.xy_surface_vertices(
                        polygon,
                        polygon_z=z,
                        exterior=False,
                        interior_index=interior_index,
                    )
                )
                for interior_index in range(len(polygon.interiors))
            ]
            if interiors:
                exterior = self.model.occ.cut(
                    [(2, exterior)],
                    [(2, interior) for interior in interiors],
                    removeObject=True,
                    removeTool=True,
                )
                self.model.occ.synchronize()
                exterior = exterior[0][0][1]  # Parse `outDimTags', `outDimTagsMap'
            surfaces.append(exterior)
        return surfaces


if __name__ == "__main__":
    # Create ring
    from shapely.geometry import Point, box
    from meshwell.model import Model

    inner_radius = 3
    outer_radius = 5

    inner_circle = Point(0, 0).buffer(inner_radius)
    outer_circle = Point(0, 0).buffer(outer_radius)
    opening = box(minx=-2, miny=-6, maxx=2, maxy=-2)

    ring = outer_circle.difference(inner_circle)  # .difference(opening)

    from shapely.plotting import plot_polygon
    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax = fig.add_subplot()
    # plot_polygon(outer_circle, ax=ax, add_points=False, color="blue")
    # plot_polygon(inner_circle, ax=ax, add_points=False, color="green")
    plot_polygon(ring, ax=ax, add_points=False, color="red")
    plt.show()

    # Test the Prism class
    polygons = ring  # Add your polygons here
    buffers = {0: 0, -1: 0, -1.001: 1, -5: 0}  # Add your buffers here

    model = Model()

    poly3D = Prism(polygons=polygons, buffers=buffers, model=model)

    model.mesh(entities_dict={"poly3D": poly3D}, filename="mesh3D.msh")



from meshwell.prism import Prism

from shapely import Polygon, box

from meshwell.model import Model

if __name__ == "__main__":
    model = Model()

    """
    DEFINE ENTITIES
    """

    simulation_width = 1
    simulation_length = 1

    substrate_thickness = 1
    cladding_thickness = 1
    pml_thickness = 1

    device_thickness = 0.1
    device_polygon = Polygon(
        shell=((0.25, 0.25), (0.25, 0.75), (0.75, 0.75), (0.25, 0.25))
    )
    device_buffers = {
        0: 0.0,
        device_thickness: 0.0,
    }

    device = Prism(
        polygons=device_polygon,
        buffers={
            0: 0.0,
            device_thickness: 0.0,
        },
        model=model,
        physical_name="device",
        mesh_order=1,
        resolution={"resolution": 0.05, "SizeMax": 1.0, "DistMax": 1.0},
    )

    background_polygon = box(
        xmin=0, ymin=0, xmax=simulation_length, ymax=simulation_width
    )

    substrate = Prism(
        polygons=background_polygon,
        buffers={
            -substrate_thickness: 0.0,
            0: 0.0,
        },
        model=model,
        physical_name="substrate",
        mesh_order=2,
    )

    cladding = Prism(
        polygons=background_polygon,
        buffers={
            0: 0.0,
            cladding_thickness: 0.0,
        },
        model=model,
        physical_name="cladding",
        mesh_order=3,
        resolution={"resolution": 0.5, "SizeMax": 1.0, "DistMax": 1.0},
    )

    PML_cladding = Prism(
        polygons=background_polygon,
        buffers={
            cladding_thickness: 0.0,
            cladding_thickness + pml_thickness: 0.0,
        },
        model=model,
        physical_name="PML_cladding",
        mesh_order=1,
    )

    PML_substrate = Prism(
        polygons=background_polygon,
        buffers={
            -substrate_thickness: 0.0,
            -substrate_thickness - pml_thickness: 0.0,
        },
        model=model,
        physical_name="PML_substrate",
        mesh_order=1,
    )

    """
    BOUNDARIES
    """

    right_polygon = box(
        xmin=simulation_length,
        ymin=0,
        xmax=simulation_length + 1,
        ymax=simulation_width,
    )
    right = Prism(
        polygons=right_polygon,
        buffers={
            -substrate_thickness - pml_thickness: 0.0,
            cladding_thickness + pml_thickness: 0.0,
        },
        model=model,
        physical_name="right",
        mesh_bool=False,
        mesh_order=0,
    )

    left_polygon = box(xmin=-1, ymin=0, xmax=0, ymax=simulation_width)
    left = Prism(
        polygons=left_polygon,
        buffers={
            -substrate_thickness - pml_thickness: 0.0,
            cladding_thickness + pml_thickness: 0.0,
        },
        model=model,
        physical_name="left",
        mesh_bool=False,
        mesh_order=0,
    )

    up_polygon = box(
        xmin=0, ymin=simulation_width, xmax=simulation_length, ymax=simulation_width + 1
    )
    up = Prism(
        polygons=up_polygon,
        buffers={
            -substrate_thickness - pml_thickness: 0.0,
            cladding_thickness + pml_thickness: 0.0,
        },
        model=model,
        physical_name="up",
        mesh_bool=False,
        mesh_order=0,
    )

    down_polygon = box(xmin=0, ymin=-1, xmax=simulation_length, ymax=0)
    down = Prism(
        polygons=down_polygon,
        buffers={
            -substrate_thickness - pml_thickness: 0.0,
            cladding_thickness + pml_thickness: 0.0,
        },
        model=model,
        physical_name="down",
        mesh_bool=False,
        mesh_order=0,
    )

    """
    ASSEMBLE AND NAME ENTITIES
    """
    entities = [
        substrate,
        device,
        cladding,
        PML_cladding,
        PML_substrate,
        up,
        down,
        left,
        right,
    ]

    mesh = model.mesh(
        entities_list=entities,
        verbosity=0,
        filename="mesh.msh",
        periodic_entities=[
            (x + "___" + s1, x + "___" + s2)
            for x in ("cladding", "substrate", "PML_cladding", "PML_substrate")
            for (s1, s2) in (("left", "right"), ("up", "down"))
        ],
    )



from meshwell.prism import Prism

from shapely import Polygon, box
from shapely.geometry import Point, MultiPolygon

from meshwell.model import Model

if __name__ == "__main__":
    model = Model()

    """
    DEFINE ENTITIES
    """

    core_thickness = 0.22
    core_width = 0.5
    core_length = 5
    box_thickness = 1
    cladding_thickness = 1

    simulation_width = 10
    simulation_length = 10
    pml_thickness = 1
    simulation_height = box_thickness + core_thickness + cladding_thickness

    # Holes
    num_holes = 5
    hole_radius = 0.1  # Define the radius of the holes
    hole_separation = 0.3  # Define the separation between the holes
    hole_offset = (
        -(num_holes - 1) * (2 * hole_radius + hole_separation) / 2 - hole_radius
    )

    # Define the centers of the num_holes holes
    hole_centers = [(i * hole_separation + hole_offset, 0) for i in range(num_holes)]

    # Create the holes as circles
    holes = [Point(center).buffer(hole_radius) for center in hole_centers]

    # Create a multipolygon from the holes
    holes_multipolygon = MultiPolygon(holes)

    # Core
    core_polygon = Polygon(
        shell=(
            (-core_length / 2, -core_width / 2),
            (core_length / 2, -core_width / 2),
            (core_length / 2, core_width / 2),
            (-core_length / 2, core_width / 2),
        ),
    )
    core_buffers = {
        0: 0.0,
        core_thickness: 0.0,
    }
    core_prism = Prism(
        polygons=core_polygon - holes_multipolygon,
        buffers=core_buffers,
        model=model,
        physical_name="core",
        mesh_order=1,
        resolution={"resolution": 0.05, "DistMax": 1.0, "SizeMax": 1.0},
    )

    cladding_polygon = box(
        xmin=-simulation_length / 2,
        ymin=-simulation_width / 2,
        xmax=simulation_length / 2,
        ymax=simulation_width / 2,
    )

    box_prism = Prism(
        polygons=cladding_polygon,
        buffers={
            -box_thickness: 0.0,
            0: 0.0,
        },
        model=model,
        physical_name="box",
        mesh_order=3,
    )

    clad_prism = Prism(
        polygons=cladding_polygon,
        buffers={
            0: 0.0,
            cladding_thickness: 0.0,
        },
        model=model,
        physical_name="cladding",
        mesh_order=2,
    )

    pml_polygon = box(
        xmin=-simulation_length / 2 - pml_thickness,
        ymin=-simulation_width / 2 - pml_thickness,
        xmax=simulation_length / 2 + pml_thickness,
        ymax=simulation_width / 2 + pml_thickness,
    )

    pml_prism = Prism(
        polygons=pml_polygon,
        buffers={
            -box_thickness - pml_thickness: 0.0,
            cladding_thickness + pml_thickness: 0.0,
        },
        model=model,
        physical_name="PML",
        mesh_order=4,
    )

    """
    ASSEMBLE AND NAME ENTITIES
    """

    mesh = model.mesh(
        entities_list=[core_prism, box_prism, clad_prism, pml_prism],
        verbosity=0,
        filename="mesh.msh",
        default_characteristic_length=0.5,
    )
